/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under Perforce control
	$Id: //depot/express/fcs992/compat/modules/converts/fld2ucd.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/db.h>
#include <avs/fld.h>
#include <avs/avs.h>
#include <avs/ucd_defs.h>

#define ERR_RETURN(A) ERRerror("fld_to_ucd",1,ERR_ORIG, A); \
                      if (*out) { \
		         UCDstructure_free(*out); \
		         *out = NULL; \
	              } \
                      return(0);
#define MAX_NAME_SIZE 1024
#define MAX_CELL   8

static char *set_names[]={"Point", "Line", "Tri", "Quad", "Tet", "Pyr",
			  "Prism", "Hex"};
static int  ucd_types[] = {UCD_POINT, UCD_LINE, UCD_TRIANGLE,
			   UCD_QUADRILATERAL, UCD_TETRAHEDRON,
		           UCD_PYRAMID, UCD_PRISM, UCD_HEXAHEDRON};
static char *cell_type[] = {"pt", "line", "tri", "quad", "tet",
			    "pyr", "prism", "hex"};

int ACfld_to_ucd PTARGS((OMobj_id elem_id));

static int fld_to_ucd(OMobj_id in, UCD_structure **out);


int
ACfld_to_ucd(OMobj_id elem_id)
{
	UCD_structure *ucd;
	OMobj_id in, out;
	int stat;
	
	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	
	stat = OMget_ptr_val(out, (void *)&ucd, 0);
	if (stat == -1) {
		ERRerror("fld_to_ucd",1,ERR_ORIG, "cannot get ucd pointer");
		return(0);
	}
	else if (stat == 0)
		ucd = NULL;

	stat = fld_to_ucd (in, &ucd);

	/* set the UCD_structure from the src OMobj */
	stat = OMset_ptr_val(out, (void *)ucd, 0);
	if (stat != 1) {
		ERRerror("fld_to_ucd",1,ERR_ORIG, "cannot set ucd pointer");
		return(0);
	}
	return(1);
}

static int
fld_to_ucd(OMobj_id in, UCD_structure **out)
{
	OMobj_id e, cell_set;
	int  i, n, size, stat, off, count;
	int  nnodes, nspace, ncomp, ncells, cell_nnodes, tot_ncells, nsets,
	     dtype, veclen, mat_id; 
	int  node_len, cell_len, cell_ncomp, node_ncomp, cs, cell;
	int  cell_tsize, node_csize, ucd_flags, nprop, *node_list, *comp_list,
	     ucd_cell_type;
	float *xc, *yc, *zc, *xfm, xform[4][4], *props;
	float xyz3[3], *rxyz, *coord, *points, min_ext[3], max_ext[3];
	float *node_data, *cell_data, *data, *rmin, *rmax, dmin, dmax;
	char  label[UCD_LABEL_LEN], labels[UCD_LABEL_LEN];
	char  unit[UCD_LABEL_LEN], units[UCD_LABEL_LEN];
	double min, max;
	
	/* Free old output data */	
	if (*out) { 
		UCDstructure_free(*out);
		*out = NULL;
	}
	
	/*  Fill in avs ucd template */
	
	stat = FLDget_xform(in, (float *)xform);
	if (stat < 0) {
		ERR_RETURN("cannot get xform for ucd");
	}
	else if (stat == 0) {
		xfm = (float *)0;
	}
	else if (MATmat_is_identity(xform, 4))
		xfm = (float *)0;
	else
		xfm = (float *)xform;
	
	if (FLDget_nnodes (in, &nnodes) != 1) {
		ERR_RETURN("Error getting nspace");
	}
	if (FLDget_nspace (in, &nspace) != 1) {
		ERR_RETURN("Error getting nspace");
	}
	
	if (FLDget_node_data_ncomp(in, &ncomp) != 1) 
		ncomp = 0;
	
	for (node_len = 0, n=0; n<ncomp; n++) {
		if (FLDget_node_data_veclen(in, n, &veclen) != 1) {
			ERR_RETURN("Error getting veclen");
		}
		node_len += veclen;
	}
	node_ncomp = ncomp;
	if (FLDget_ncell_sets(in, &nsets) != 1) {
		ERR_RETURN("cannot get nsets");
	}
	
	for (tot_ncells=0, cell_tsize=0, cs=0; cs<nsets; cs++) {
		if (FLDget_cell_set(in, cs, &cell_set) != 1) {
			ERR_RETURN("cannot get cell set");
		}
		if (FLDget_ncells(cell_set, &ncells) != 1) {
			ERR_RETURN("cannot get ncells");
		}
		tot_ncells += ncells;
		
		if (FLDget_cell_set_nnodes(cell_set,  &cell_nnodes) != 1) {
			ERR_RETURN("cannot get cell nnodes");
		}
		cell_tsize += ncells*cell_nnodes;
		
		if (FLDget_cell_data_ncomp (cell_set, &ncomp) != 1) {
			ncomp = 0;
		}
		if (cs == 0) {
			for (cell_len = 0, n=0; n<ncomp; n++) {
				if (FLDget_cell_data_veclen(cell_set, n,
							    &veclen) != 1) {
					ERR_RETURN("Error getting veclen");
				}
				cell_len += veclen;
			}
			cell_ncomp = ncomp;
		}
		else {
			if (ncomp != cell_ncomp) {
				ERR_RETURN(
				"cell sets with different data, cannot convert");
			}
		}
		if (FLDget_cell_nprops (cell_set, &nprop) != 1) {
			nprop = 0;
		}
	}
	
	ucd_flags = UCD_INT;
	if (nprop)
		ucd_flags |= UCD_MATERIAL_IDS;
	
	node_csize = 0;

	/********************************************/
	/* Allocate ucd structure                   */
	/********************************************/
	*out =(UCD_structure *)UCDstructure_alloc ("converted", 0, 
			   ucd_flags, tot_ncells,  cell_tsize, cell_len,
			   nnodes, node_csize, node_len, 0); 

	if (*out == NULL) {
		ERR_RETURN("cannot allocate AVS ucd");
	}


	/********************************************/
	/* Set the extents                          */
	/********************************************/
	for (i=0; i<3; i++) {
		min_ext[i]=0.0;
		max_ext[i]=0.0;
	}
	if (FLDget_coord_extent(in, min_ext, max_ext) != 1) {
		ERR_RETURN("cannot get extents");
	}
	if (xfm) {
		MATvec3_mat4_multiply(min_ext, xfm);
		MATvec3_mat4_multiply(max_ext, xfm);
	}		

	UCDstructure_set_extent (*out, min_ext, max_ext);

	/********************************************/
	/* Set the coordinates                      */
	/********************************************/
	if (FLDget_coord(in,(float **)&coord, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("Error getting coordinates");
	}
	UCDstructure_get_node_positions (*out, &xc, &yc, &zc);

	for (i=0; i<nnodes; i++) {
		xyz3[0]=xyz3[1]=xyz3[2]=0.0;
		memcpy((void *)xyz3, (void *)(coord+nspace*i),
		       nspace*sizeof(float));
		if (xfm)
			MATvec3_mat4_multiply(xyz3, xfm);
		xc[i] = xyz3[0];
		yc[i] = xyz3[1];
		zc[i] = xyz3[2];
	}
		ARRfree((char *)coord);

	/********************************************/
	/* Set node data                            */
	/********************************************/
	if (node_len) {
		UCDstructure_get_node_data (*out, &node_data);
		dtype = DTYPE_FLOAT;
		comp_list = (int *)malloc(sizeof(int) * node_len);
		rmin = (float *)malloc(sizeof(float) * node_len);
		rmax = (float *)malloc(sizeof(float) * node_len);
		if (rmin == NULL || rmax == NULL || comp_list == NULL) {
			ERR_RETURN("cannot alloc minmax");
		}
		for (off=0, i=0; i<node_ncomp; i++) {
			if (FLDget_node_data_label(in, i, label,
						   UCD_LABEL_LEN) != 1) {
				sprintf(label, "data_%d", i);
			}
			if (i==0) 
				strcpy(labels, label);
			else {
				strcat(labels, ".");
				strcat(labels, label);
			}
			if (FLDget_node_data_units(in, i, unit,
						   UCD_LABEL_LEN) != 1) {
				strcpy(unit,"x");
			}
			if (i==0) 
				strcpy(units, unit);
			else {
				strcat(units, ".");
				strcat(units, unit);
			}
			if (FLDget_node_data_veclen(in, i, &veclen) != 1) {
				ERR_RETURN("Error getting veclen");
			}
			comp_list[i] = veclen;
			
			if (FLDget_typed_node_data(in, i, &dtype,
						   (char **)&data, &size,
						   OM_GET_ARRAY_RD) != 1) {
				ERR_RETURN("Error getting node_datas");
			}
			memcpy((void *)(node_data+off),
			       (void *)data, nnodes*veclen*sizeof(float));
			off += nnodes*veclen;
			
			if (FLDget_node_data_minmax(in, i, (char *)&min,
						    (char *)&max) != 1) {
				ERR_RETURN("Error getting node minmax");
			}
			UTILtype_to_float(rmin+i, (char *)&min, dtype);
			UTILtype_to_float(rmax+i, (char *)&max, dtype);
			ARRfree((char *)data);
		}
		UCDstructure_set_node_components (*out, comp_list, node_ncomp);
		UCDstructure_set_node_labels(*out, labels, ".");
		UCDstructure_set_node_units(*out, units, ".");
		UCDstructure_set_node_minmax (*out, rmin, rmax);
		free(rmin);
		free(rmax);
		free(comp_list);
	}
	if (cell_len) {
		UCDstructure_get_cell_data (*out, &cell_data);
		dtype = DTYPE_FLOAT;
		comp_list = (int *)malloc(sizeof(int) * cell_len);
		rmin = (float *)malloc(sizeof(float) * cell_len);
		rmax = (float *)malloc(sizeof(float) * cell_len);
		if (rmin == NULL || rmax == NULL || comp_list == NULL) {
			ERR_RETURN("cannot alloc minmax");
		}
	}
	for (count=0,cs=0; cs<nsets; cs++) {
		if (FLDget_cell_set(in, cs, &cell_set) != 1) {
			ERR_RETURN("cannot get cell set");
		}
		if (FLDget_ncells(cell_set, &ncells) != 1) {
			ERR_RETURN("cannot get ncells");
		}
		if (FLDget_cell_set_name(cell_set, label, UCD_LABEL_LEN) != 1) {
			ERR_RETURN("cannot get cell name");
		}
		ucd_cell_type = -1;
		for (i=0; i< MAX_CELL; i++) {
			if (strcmp(label, set_names[i]) == 0) {
				ucd_cell_type = ucd_types[i];
				break;
			}
		}
		if (ucd_cell_type < 0) {
			ERR_RETURN("unknown cell type, cannot convert");
		}
		if (FLDget_cell_set_nnodes(cell_set,  &cell_nnodes) != 1) {
			ERR_RETURN("cannot get cell nnodes");
		}
		if (FLDget_node_connect(cell_set, &node_list, 
					&size, OM_GET_ARRAY_RD) != 1) {
			ERR_RETURN("cannot get cell connectivity");
		}
		if (FLDget_cell_nprops (cell_set, &nprop) != 1) {
			nprop = 0;
		}
		if (nprop) {
			if (FLDget_cell_props(cell_set, (float **)&props,
					      &size, OM_GET_ARRAY_RD) != 1) {
				ERR_RETURN("Error getting cell props");
			}
			mat_id = (int)props[0];
		}
		else
			mat_id = 0;
		for (cell=0; cell<ncells; cell++) {
			UCDcell_set_information (*out, count, (char *)(count+1),
						 cell_type[ucd_cell_type],
						 mat_id, ucd_cell_type, 0,
						 node_list+cell*cell_nnodes);
			count++;
		}
		if (cell_len) {
			for (off=0, i=0; i<cell_ncomp; i++) {
				if (FLDget_cell_data_minmax(cell_set, i,
							    (char *)&min,
							    (char *)&max) != 1)
				{
					ERR_RETURN("Error getting cell minmax");
				}
				UTILtype_to_float(&dmin, (char *)&min, dtype);
				UTILtype_to_float(&dmax, (char *)&max, dtype);

				if (FLDget_cell_data_veclen(cell_set, i,
							    &veclen) != 1) {
					ERR_RETURN("Error getting veclen");
				}
				if (cs == 0) {
					comp_list[i] = veclen;
					rmin[i] = dmin;
					rmax[i] = dmax;
					if (FLDget_cell_data_label(cell_set, i,
						   label, UCD_LABEL_LEN) != 1) {
						sprintf(label, "data_%d", i);
					}
					if (i==0) 
						strcpy(labels, label);
					else {
						strcat(labels, ".");
						strcat(labels, label);
					}
					if (FLDget_cell_data_units(cell_set, i,
						    unit, UCD_LABEL_LEN) != 1) {
						strcpy(unit,"x");
					}
					if (i==0) 
						strcpy(units, unit);
					else {
						strcat(units, ".");
						strcat(units, unit);
					}
				}
				else {
					if (veclen != comp_list[i]) {
						ERR_RETURN("cell data has different veclen, cannot convert");
					}
					if (dmin < rmin[i])
						rmin[i] = dmin;
					if (dmax > rmax[i])
						rmax[i] = dmax;
				}
			}
		}
	}
	/********************************************/
	/* Set cell data                            */
	/********************************************/
	if (cell_len) {
		for (off=0, i=0; i<cell_ncomp; i++) {
			for (cs=0; cs<nsets; cs++) {
				if (FLDget_cell_set(in, cs, &cell_set) != 1) {
					ERR_RETURN("cannot get cell set");
				}
				if (FLDget_ncells(cell_set, &ncells) != 1) {
					ERR_RETURN("cannot get ncells");
				}
				if (FLDget_typed_cell_data(cell_set, i, &dtype,
						       (char **)&data, &size,
						       OM_GET_ARRAY_RD) != 1) {
					ERR_RETURN("Error getting cell_datas");
				}
				memcpy((void *)(cell_data+off), (void *)data,
				       ncells*veclen*sizeof(float));
				off += ncells*veclen;
				
				ARRfree((char *)data);
			}
		}
			
		UCDstructure_set_cell_components (*out, comp_list, cell_ncomp);
		UCDstructure_set_cell_labels(*out, labels, ".");
		UCDstructure_set_cell_units(*out, units, ".");
		UCDstructure_set_cell_minmax (*out, rmin, rmax);
		free(rmin);
		free(rmax);
		free(comp_list);
	}
	/* Indicate success */
	return(1);
}


